export ThreadedModel

"""
    ThreadedModel
Experimental model wrapper than enables multi-threading when evaluating `intensitymap`
"""
struct ThreadedModel{M} <: AbstractModel
    model::M
end

@inline ThreadedModel(model::ThreadedModel) = model
@inline ThreadedModel(model::M) where {M<:CompositeModel} = ThreadedModel(imanalytic(M), model)
@inline ThreadedModel(::IsAnalytic, model::CompositeModel)        = ThreadedModel{typeof(model)}(model)
@inline ThreadedModel(::NotAnalytic, model::AddModel)             = AddModel(ThreadedModel(model.m1), ThreadedModel(model.m2))
@inline ThreadedModel(::NotAnalytic, model::ConvolvedModel)       = ConvolvedModel(ThreadedModel(model.m1), ThreadedModel(model.m2))
@inline ThreadedModel(model::ModelImage) = @set model.model = ThreadedModel(model.model)

Base.@constprop :aggressive @inline visanalytic(::Type{<:ThreadedModel{M}}) where {M} = visanalytic(M)
Base.@constprop :aggressive @inline imanalytic(::Type{<:ThreadedModel{M}}) where {M} = imanalytic(M)
Base.@constprop :aggressive @inline ispolarized(::Type{<:ThreadedModel{M}}) where {M} = ispolarized(M)


@inline visibility_point(m::ThreadedModel, u, v, time, freq) = visibility_point(m.model, u, v, time, freq)
@inline intensity_point(m::ThreadedModel, p) = intensity_point(m.model, p)

@inline radialextent(m::ThreadedModel) = radialextent(basemodel(m))
@inline flux(m::ThreadedModel) = flux(basemodel(m))

function intensitymap_analytic(s::ThreadedModel, g::RectiGrid)
    T = typeof(intensity_point(s, (X=g.X[begin], Y=g.Y[begin])))
    img = IntensityMap(Array{T}(undef, length(g.X), length(g.Y)), g)
    return intensitymap_analytic!(img, s)
end

# function visibilities_analytic(m::ThreadedModel, u, v, time, freq)
#     T = typeof(visibility_point(m, first(u), first(v), first(time), first(freq)))
#     vis = similar(u, T)
#     visibilities_analytic!(vis, m, u, v, time, freq)
#     return vis
# end

# function visibilities_analytic!(vis::AbstractArray, m::ThreadedModel, u, v, time, freq)
#     mbase = m.model
#     Threads.@threads for i in eachindex(vis, u, v, time, freq)
#         vis[i] = visibility_point(mbase, u[i], v[i], time[i], freq[i])
#     end
#     return nothing
# end

intensitymap(m, p, threaded::Bool) = intensitymap(m, p, static(threaded))
intensitymap(m, p, ::False) = intensitymap(m, p)
intensitymap(m, p, ::True)  = intensitymap(ThreadedModel(m), p)

intensitymap!(img::IntensityMapTypes, m, threaded::Bool) = intensitymap!(img, m, static(threaded))
intensitymap!(img::IntensityMapTypes, m, ::False) = intensitymap!(img, m)
intensitymap!(img::IntensityMapTypes, m, ::True)  = intensitymap!(img, ThreadedModel(m))

function intensitymap!(::IsAnalytic, img::IntensityMap, s::ThreadedModel)
    dx, dy = pixelsizes(img)
    mm = Base.Fix1(intensity_point, s)
    g = imagegrid(img)
    Threads.@threads for I in CartesianIndices(img)
        @inbounds img[I] = mm(g[I])*dx*dy
    end
    return img
end

function fouriermap(::IsAnalytic, m::ThreadedModel, dims::ComradeBase.AbstractGrid)
    X = dims.X
    Y = dims.Y
    uu,vv = uviterator(length(X), step(X), length(Y), step(Y))
    # uvgrid = ComradeBase.grid(U=uu, V=vv)
    T = typeof(visibility_point(m, uu[1], vv[1], 0, 0))
    vis = similar(uu, T, (length(uu), length(vv)))
    Threads.@threads for I in CartesianIndices(vis)
        ix, iy = Tuple(I)
        vis[I] = visibility_point(m, uu[ix], vv[iy], 0, 0)
    end
    return IntensityMap(vis, (U=uu, V=vv))
end
